# -*- coding: UTF-8 -*-
require 'AlmanacDateTime'
require 'AlmanacSolarTerm'
require 'AlmanacMath'

#
#= 干支（Oriental Zodiac または Chinese zodiac）算出用クラスライブラリ
#
#Authors::   Yoshihiko Hara(GoroNeko)
#Date::      2010/08/22
#Version::   0.9.0
#Copyright:: Copyright (C) Yoshihiko Hara(GoroNeko), 2010. All rights reserved.
#License::   GNU GENERAL PUBLIC LICENSE Version 2
#

class AlmanacOrientalZodiac

  # 本クラスライブラリの、名前とバージョン
  NAME = "AlmanacOrientalZodiac"
  VERSION = "0.9.0"

  # 干支(Oriental Zodiac)を示す
  # 0(癸亥)～59(壬戌)
  ORIENTAL_ZODIAC_NUMBER \
    = Array.new([0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                 20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
                 30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
                 40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
                 50, 51, 52, 53, 54, 55, 56, 57, 58, 59])

  # 時の干支番号算出用テーブル
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB0 \
    = Array.new([1,13,25,37,49])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB1 \
    = Array.new([2,14,26,38,50])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB2 \
    = Array.new([3,15,27,39,51])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB3 \
    = Array.new([4,16,28,40,52])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB4 \
    = Array.new([5,17,29,41,53])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB5 \
    = Array.new([6,18,30,42,54])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB6 \
    = Array.new([7,19,31,43,55])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB7 \
    = Array.new([8,20,32,44,56])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB8 \
    = Array.new([9,21,33,45,57])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB9 \
    = Array.new([10,22,34,46,58])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB10 \
    = Array.new([11,23,35,47,59])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB11 \
    = Array.new([12,24,36,48,0])
  ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME \
    = Array.new([ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB0,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB1,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB2,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB3,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB4,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB5,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB6,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB7,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB8,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB9,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB10,
                 ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME_SUB11])
  # 時の干支番号算出用第１キー格納テーブル
  FIRST_KEY_OF_ORIENTAL_ZODIAC_FOR_TIME \
    = Array.new([0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,0])
  # 時の干支番号算出用第２キー格納テーブル
  SECOND_KEY_OF_ORIENTAL_ZODIAC_FOR_TIME \
      = Array.new([4,0,1,2,3,4,0,1,2,3])
  # 干支の漢字和名
  ORIENTAL_ZODIAC_NAME_J \
    = Array.new(["癸亥", "甲子", "乙丑", "丙寅", "丁卯", "戊辰",
                 "己巳", "庚午", "辛未", "壬申", "癸酉", "甲戌",
                 "乙亥", "丙子", "丁丑", "戊寅", "己卯", "庚辰",
                 "辛巳", "壬午", "癸未", "甲申", "乙酉", "丙戌",
                 "丁亥", "戊子", "己丑", "庚寅", "辛卯", "壬辰",
                 "癸巳", "甲午", "乙未", "丙申", "丁酉", "戊戌",
                 "己亥", "庚子", "辛丑", "壬寅", "癸卯", "甲辰",
                 "乙巳", "丙午", "丁未", "戊申", "己酉", "庚戌",
                 "辛亥", "壬子", "癸丑", "甲寅", "乙卯", "丙辰",
                 "丁巳", "戊午", "己未", "庚申", "辛酉", "壬戌"])
  # 干支の英名
  # ウィキペディア英語版の記事を参考にした
  # (http://en.wikipedia.org/wiki/Chinese_zodiac)
  ORIENTAL_ZODIAC_NAME_E \
    = Array.new(["Yin Water Boar", "Yang Wood Rat", "Yin Wood Ox", "Yang Water Tiger", "Yin Earth Rabbit", "Yang Fire Dragon",
                 "Yin Earth Snake", "Yang Metal Horse", "Yin Metal Sheep", "Yang Water Monkey", "Yin Water Rooster", "Yang Wood Dog",
                 "Yin Wood Boar", "Yang Water Rat", "Yin Earth Ox", "Yang Fire Tiger", "Yin Earth Rabbit", "Yang Metal Dragon",
                 "Yin Metal Snake", "Yang Water Horse", "Yin Water Sheep", "Yang Wood Monkey", "Yin Wood Rooster", "Yang Water Dog",
                 "Yin Earth Boar", "Yang Fire Rat", "Yin Earth Ox", "Yang Metal Tiger", "Yin Metal Rabbit", "Yang Water Dragon",
                 "Yin Water Snake", "Yang Wood Horse", "Yin Wood Sheep", "Yang Water Monkey", "Yin Earth Rooster", "Yang Fire Dog",
                 "Yin Earth Boar", "Yang Metal Rat", "Yin Metal Ox", "Yang Water Tiger", "Yin Water Rabbit", "Yang Wood Dragon",
                 "Yin Wood Snake", "Yang Water Horse", "Yin Earth Sheep", "Yang Fire Monkey", "Yin Earth Rooster", "Yang Metal Dog",
                 "Yin Metal Boar", "Yang Water Rat", "Yin Water Ox", "Yang Wood Tiger", "Yin Wood Rabbit", "Yang Water Dragon",
                 "Yin Earth Snake", "Yang Fire Horse", "Yin Earth Sheep", "Yang Metal Monkey", "Yin Metal Rooster", "Yang Water Dog"])
  # 十干を示す
  # 0(癸)～9(壬)
  JIKKAN_NUMBER \
    = Array.new([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
  # 十干の漢字和名
  JIKKAN_NAME_J \
    = Array.new(["癸", "甲", "乙", "丙", "丁", "戊", "己", "庚", "辛", "壬"])
  # 十干の英名
  # ウィキペディア英語版の記事を参考にした
  # (http://en.wikipedia.org/wiki/Chinese_zodiac)
  JIKKAN_NAME_E \
    = Array.new(["Yin Water", "Yang Wood", "Yin Wood", "Yang Water", "Yin Earth", 
                 "Yang Fire", "Yin Earth", "Yang Metal", "Yin Metal", "Yang Water"])
  # 十二支を示す
  # 0(亥)～11(戌)
  JYUUNISHI_NUMBER \
    = Array.new([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
  # 十二支の漢字和名
  JYUUNISHI_NAME_J \
    = Array.new(["亥","子","丑","寅","卯","辰",
                 "巳","午","未","申","酉","戌"])
  # 十二支の英名
  # ウィキペディア英語版の記事を参考にした
  # (http://en.wikipedia.org/wiki/Chinese_zodiac)
  JYUUNISHI_NAME_E \
    = Array.new(["Boar", "Rat", "Ox", "Tiger", "Rabbit", "Dragon",
                 "Snake", "Horse", "Sheep", "Monkey", "Rooster", "Dog"])

  # 指定西暦年月日(oriental_zodiac_date_time:AlmanacDateTime)の
  # 年の干支番号(oriental_zodiac_of_year_number：Integer)を返す
  #
  # 【引数】
  #   1.指定西暦年月日(oriental_zodiac_date_time)
  # 【戻り値】
  #   1.年の干支番号(oriental_zodiac_of_year_number)
  #
  def get_oriental_zodiac_of_year_number(oriental_zodiac_date_time)

    # 指定年月日の時差を調整し、日本における日時に調整しておく
    oriental_zodiac_date_time \
      = oriental_zodiac_date_time.change_offset_to_japan(oriental_zodiac_date_time)

    target_year = oriental_zodiac_date_time.year

    # 年の干支は立春をもって切り替わる。
    # 月日がその年の立春を超えていない（月が1月、または、2月かつ立春の日より前の日付）場合、
    # 前年の年の干支番号を求める。

    # その年の立春の日付を得る。
    # 立春は2月の"節"(節気番号 = 0)
    rissyun_date_time = \
      AlmanacSolarTerm.new.get_sekki_date_time(oriental_zodiac_date_time.year, \
                                             oriental_zodiac_date_time.offset, \
                                             0)

    if 1 == oriental_zodiac_date_time.month
      target_year -= 1
    end

    if 2 == oriental_zodiac_date_time.month && \
       rissyun_date_time.day > oriental_zodiac_date_time.day
      target_year -= 1
    end

    # 修正後の年日時から得た年情報をもとに年の干支番号を求める
    # 年の干支は60年で一巡することから、年の干支番号は以下の式で求められる
    # 年の干支番号 = 西暦年数を60で割った余り
    oriental_zodiac_of_year_number = (target_year % 60) - 3

    # 年の干支番号を0～59の範囲におさめるための補正を行う。
    if 0 > oriental_zodiac_of_year_number
      oriental_zodiac_of_year_number += 60
    end

    return oriental_zodiac_of_year_number
  end

  # 指定西暦年月日(oriental_zodiac_date_time:AlmanacDateTime)の
  # 月の干支番号(oriental_zodiac_of_month_number：Integer)を返す
  #
  # 【引数】
  #   1.指定西暦年月日(oriental_zodiac_date_time)
  # 【戻り値】
  #   1.月の干支番号(oriental_zodiac_of_month_number)
  #
  def get_oriental_zodiac_of_month_number(oriental_zodiac_date_time)

    # 指定年月日の時差を調整し、日本における日時に調整しておく
    oriental_zodiac_date_time \
      = oriental_zodiac_date_time.change_offset_to_japan(oriental_zodiac_date_time)

    oriental_zodiac_of_month_number = 0
    if 1900 > oriental_zodiac_date_time.year
      oriental_zodiac_of_month_number = \
        60 - (((13 - oriental_zodiac_date_time.month) \
        + ((1900 - (oriental_zodiac_date_time.year + 1)) * 12)) % 60)
    else
      oriental_zodiac_of_month_number = \
        ((oriental_zodiac_date_time.month - 1) \
        + ((oriental_zodiac_date_time.year - 1900) * 12)) % 60
    end

    oriental_zodiac_of_month_number += 14

    # 月の干支は、その月の節日をもって切り替わる。
    # 月日がその月の節日を超えていない場合、前月の年の干支番号を求める
    # （求める月の干支番号 = ここまでで求まった月の干支番号 - 1）。

    # その月の節日の日付を得る。
    sekki_number = 0
    if 1 == oriental_zodiac_date_time.month
      sekki_number = 22
    else
      sekki_number = (oriental_zodiac_date_time.month - 2) * 2
    end

    setsu_date_time \
      = AlmanacSolarTerm.new.get_sekki_date_time(oriental_zodiac_date_time.year, \
                                               oriental_zodiac_date_time.offset, \
                                               sekki_number)

    if setsu_date_time.day > oriental_zodiac_date_time.day
      oriental_zodiac_of_month_number -= 1
    end

    # 月の干支番号を0～59の範囲におさめるための補正を行う。
    if oriental_zodiac_of_month_number > 59
      oriental_zodiac_of_month_number -= 60
    end

    return oriental_zodiac_of_month_number
  end

  # 指定西暦年月日(oriental_zodiac_date_time:AlmanacDateTime)の
  # 日の干支番号(oriental_zodiac_of_day_number：Integer)を返す
  #
  # 【引数】
  #   1.指定西暦年月日(oriental_zodiac_date_time)
  # 【戻り値】
  #   1.日の干支番号(oriental_zodiac_of_day_number)
  #
  def get_oriental_zodiac_of_day_number(oriental_zodiac_date_time)

    # 指定年月日の時差を調整し、日本における日時に調整しておく
    oriental_zodiac_date_time \
      = oriental_zodiac_date_time.change_offset_to_japan(oriental_zodiac_date_time)

    # 指定西暦年月日（AlmanacDateTime）の
    # ユリウス通日（ajd）から日の干支番号を得る
    oriental_zodiac_of_day_number \
      = AlmanacMath.new.fmod((oriental_zodiac_date_time.ajd - 9.0), 60.0).to_i

    return oriental_zodiac_of_day_number
  end

  # 指定西暦年月日(oriental_zodiac_date_time:AlmanacDateTime)の
  # 時の干支番号(oriental_zodiac_of_hour_number：Integer)を返す
  #
  # 【引数】
  #   1.指定西暦年月日(oriental_zodiac_date_time)
  # 【戻り値】
  #   1.時の干支番号(oriental_zodiac_of_hour_number)
  #
  def get_oriental_zodiac_of_hour_number(oriental_zodiac_date_time)

    # 指定年月日の時差を調整し、日本における日時に調整しておく
    oriental_zodiac_date_time.offset \
      oriental_zodiac_date_time.change_offset_to_japan(oriental_zodiac_date_time)

    # 指定西暦年月日の時間（oriental_zodiac_date_time.hour）を第１キーとして
    # 指定西暦年月日の日の十干番号（jikkann_number）を第２キーとして
    # 時の干支番号(oriental_zodiac_of_hour_number：Integer)を求める
    first_key = FIRST_KEY_OF_ORIENTAL_ZODIAC_FOR_TIME[oriental_zodiac_date_time.hour]
    second_key \
      = SECOND_KEY_OF_ORIENTAL_ZODIAC_FOR_TIME[get_jikkann_number( \
          get_oriental_zodiac_of_day_number(oriental_zodiac_date_time))]
    
    oriental_zodiac_of_hour_number \
      = ORIENTAL_ZODIAC_NUMBER_TABLE_FOR_TIME[first_key][second_key]

    return oriental_zodiac_of_hour_number
  end

  # 指定干支番号(oriental_zodiac_number：Integer)に対応する
  # 十干番号(jikkann_number：Integer)を返す
  #
  # 【引数】
  #   1.指定干支番号(oriental_zodiac_number)
  # 【戻り値】
  #   1.十干番号(jikkann_number：Integer)
  #
  def get_jikkann_number(oriental_zodiac_number)
    
    # 十干番号 = 干支番号 % 10
    jikkann_number = oriental_zodiac_number % 10

    return jikkann_number
  end

  # 指定干支番号(oriental_zodiac_number：Integer)に対応する
  # 十二支番号(jyuunishi_number：Integer)を返す
  #
  # 【引数】
  #   1.指定干支番号(oriental_zodiac_number)
  # 【戻り値】
  #   1.十二支番号(jyuunishi_number：Integer)
  #
  def get_jyuunishi_number(oriental_zodiac_number)
    
    # 十二支番号 = 干支番号 % 12
    jyuunishi_number = oriental_zodiac_number % 12

    return jyuunishi_number
  end

end
